//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		map_misc.c
//
// purpose	Define miscellaneous functions.
//-----------------------------------------------------------------------------

#include "map_lib.h"

//-----------------------------------------------------------------------------
// struct	map_link
// struct	map_node
//
// purpose	Define structures needed to keep the AVL trees used to hold
//		the mapping.
//-----------------------------------------------------------------------------
struct map_link {
    avl_link		link    ;
    // The key is found here in the type specific structs.
};
struct map_node {
    struct map_mixed	data    ;
    struct map_link	link	;
};

//-----------------------------------------------------------------------------
// function	map_free_node (internal)
//
// purpose	Free a node, including data content where applicable.
//
// arguments	1 (void *) pointer to variant node type
//
// returns	(void)
//-----------------------------------------------------------------------------
static
void
map_free_node (
    void *	arg_node
)
{
    struct map_node *	node_p	;

    node_p = arg_node;

    //-- Free the data space that was used.
    if ( map_type_is_string( node_p->data.type ) && node_p->data.mix.string.dup ) {
	free( node_p->data.mix.string.ptr );
    }

    //-- Free the node itself, with key.
    free( node_p );

    return;
}

//-----------------------------------------------------------------------------
// function	map_key_type
//
// purpose	Return the key type of the mapping.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) >0 = key type returned
//-----------------------------------------------------------------------------
int
map_key_type (
    MAP		arg_map
)
{
    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Return the key type.
    return arg_map->key_type;
}

//-----------------------------------------------------------------------------
// function	map_data_type
//
// purpose	Return the data type of the active member of a mapping.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no active member
//		(int)  0 = member found, no data
//		(int) >0 = member found, data type returned
//-----------------------------------------------------------------------------
int
map_data_type (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Get the active node and check it.
    node_p = avl_find( & (arg_map->avl_tree), NULL, 0 );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_data_clear
// alias	map_none_store
//
// purpose	Clear the data value of the active member of a mapping so
//		that it is just like it was newly inserted.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no current member to clear
//		(int)  0 = current member cleared OK
//-----------------------------------------------------------------------------
int
map_data_clear (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Get the active node and check it.
    node_p = avl_find( & (arg_map->avl_tree), NULL, 0 );
    if ( ! node_p ) return -1;

    //-- Free the data space that was used.
    if ( map_type_is_string( node_p->data.type ) && node_p->data.mix.string.dup ) {
	free( node_p->data.mix.string.ptr );
    }

    //-- Mark the node has having no data.
    node_p->data.type = 0;

    //-- Status is OK.
    return 0;
}


//-----------------------------------------------------------------------------
// function	map_first
//
// purpose	Select the first member of a mapping as the active one.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no members in map
//		(int)  0 = member found, no data
//		(int) >0 = member found, data type returned
//-----------------------------------------------------------------------------
int
map_first (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer find the first member.
    node_p = avl_first( & (arg_map->avl_tree) );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_last
//
// purpose	Select the last member of a mapping as the active one.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no members in map
//		(int)  0 = member found, no data
//		(int) >0 = member found, data type returned
//-----------------------------------------------------------------------------
int
map_last (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer find the last member.
    node_p = avl_last( & (arg_map->avl_tree) );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_forward
//
// purpose	Move the current member selection of a mapping one step in
//		the forward direction.  If there is no current member, then
//		select the first member.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no member found
//		(int)  0 = member found, no data
//		(int) >0 = member found, data type returned
//-----------------------------------------------------------------------------
int
map_forward (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer find the next forward member.
    node_p = avl_forward( & (arg_map->avl_tree), NULL );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_reverse
//
// purpose	Move the current member selection of a mapping one step in
//		the reverse direction.  If there is no current member, then
//		select the last member.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no member found
//		(int)  0 = member found, no data
//		(int) >0 = member found, data type returned
//-----------------------------------------------------------------------------
int
map_reverse (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer find the next reverse member.
    node_p = avl_reverse( & (arg_map->avl_tree), NULL );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_unselect
//
// purpose	Clear the current member selection and make the mapping have
//		no currently selected member.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no member was selected before
//		(int)  1 = an member was selected before
//-----------------------------------------------------------------------------
int
map_unselect (
    MAP		arg_map
)
{
    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer do the unselect.
    return ( avl_unselect( & (arg_map->avl_tree) ) ) ? 1 : -1;
}

//-----------------------------------------------------------------------------
// function	map_delete
//
// purpose	Delete the currently selected member of a mapping without
//		consideration of the key type.
//
// arguments	1 (MAP) mapping handle
//
// returns	(int) -2 = error, no map
//		(int) -1 = no current member to delete
//		(int)  1 = current member deleted OK
//-----------------------------------------------------------------------------
int
map_delete (
    MAP		arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer delete the member.
    if ( ! ( node_p = avl_delete( & (arg_map->avl_tree), NULL ) ) ) return -1;

    //-- Free all the space that was used for this node.
    map_free_node( node_p );

    //-- Status is OK.
    return 1;
}

//-----------------------------------------------------------------------------
// function	map_empty
//
// purpose	Delete all the members of a mapping without consideration of
//		the key type.
//
// arguments	1 (MAP) mapping handle
//
// returns	(long) -2 = error, no map
//		(long) number of members deleted
//-----------------------------------------------------------------------------
long
map_empty (
    MAP		arg_map
)
{
    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer empty the whole tree, returning the node count.
    return avl_empty( & (arg_map->avl_tree), map_free_node );
}

//-----------------------------------------------------------------------------
// function	map_destroy
//
// purpose	Destroy a mapping.  Delete all the nodes and free the mapping.
//		The mapping handle becomes invalid and must not be used again.
//
// arguments	1 (MAP) mapping handle
//
// returns	(long) -2 = error, no map
//		(long) number of members deleted
//
// note		In UNIX and similar virtual memory systems, if the process
//		is near the end of its execution, it is likely not necessary
//		to free the remaining space, as virtual memory cleanup will
//		remove it more efficiently.
//-----------------------------------------------------------------------------
long
map_destroy (
    MAP		arg_map
)
{
    long	count	;

    //-- Make sure we have a mapping.
    if ( ! arg_map ) return -2;

    //-- Let the AVL layer empty the whole tree, returning the node count.
    count = avl_empty( & (arg_map->avl_tree), map_free_node );

    //-- Free the mapping itself.
    free( arg_map );

    //-- Now we can return the count.
    return count;
}
