//-----------------------------------------------------------------------------
// Copyright  1999, 2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// program	wordfreq
//
// purpose	Count the frequency of different word items in the input.
//
//		Demonstrate how single node can be defined to have 2 tree
//		links allowing each instance to be in 2 tree classes at
//		the same time.
//
// syntax	wordfreq [-q] ["sepchars"] < infile
//-----------------------------------------------------------------------------
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <avl.h>

#define WORDSIZE	256

//-----------------------------------------------------------------------------
// Define a struct that implements a link for a character string key.
// Define a struct that implements a link for an unsigned long key.
// Define a struct that implements a node to carry both links together.
// Define a function that compares the character string keys in two links.
// Define a function that compares the unsigned long keys in two links.
//-----------------------------------------------------------------------------
struct link_str {
    avl_link		link		;
    char *		key		;
};

struct link_ul {
    avl_link		link		;
    unsigned long	key		;
};

struct node {
    struct link_str	word		;
    struct link_ul	count		;
    char		data	[16]	;
};

static int cmp_str( avl_link * arg_one, avl_link * arg_two, AVL * arg_tree )
{
    return strcmp(
	((struct link_str *)arg_one)->key,
	((struct link_str *)arg_two)->key
    );
}

static int cmp_ul( avl_link * arg_one, avl_link * arg_two, AVL * arg_tree )
{
    return ( ((struct link_str *)arg_one)->key < ((struct link_str *)arg_two)->key ) ?
	-1 : ( ((struct link_str *)arg_one)->key > ((struct link_str *)arg_two)->key ) ? 1 : 0;
}

//	static const char * key_str( void * arg_node )
//	{
//	    return ((struct node *)(arg_node))->data;
//	}

//-----------------------------------------------------------------------------
// function	readword
//-----------------------------------------------------------------------------
static inline
int
readword (
    char *	arg_buffer
,
    size_t	arg_length
,
    char *	arg_seps
)
{
    char *	p	;
    char *	q	;
    int		c	;

    //-- Note where the end of the buffer is.
    q = ( p = arg_buffer ) + arg_length - 1;

    //-- Skip all leading separator characters.
    while ( (c = getc( stdin )) != EOF && arg_seps[ c ] ) {
    }

    //-- Get characters until end of word, store until full.
    while ( c != EOF && arg_seps[ c ] == 0 ) {
	if ( p < q ) * p ++ = c;
	c = getc( stdin );
    }
    * p = 0;

    //-- Return 0 if EOF, 1 if a word found.
    return ( c == EOF );
}

//-----------------------------------------------------------------------------
// function	main
//-----------------------------------------------------------------------------
int
main (
    int		argc
,
    char * *	argv
)
{
    AVL			by_word		;
    AVL			by_count	;

    struct link_str	key_word	;

    struct node *	node_p		;
    char		word		[ WORDSIZE ];
    char		seps		[ 256 ];
    size_t		len		;
    unsigned long	reads		;
    unsigned long	words		;
    unsigned long	counts		;
    int			quiet		;


    //------------------------
    // Check for quiet option.
    //------------------------
    quiet = 0;
    if ( argc > 1 && argv[1] &&
	 argv[1][0] == '-' && argv[1][1] == 'q' && argv[1][2] == 0 ) {
	quiet = 1;
	-- argc;
	++ argv;
    }

    //------------------------------------------------------------------------
    // Initialize the separator table that is given to readword().  This table
    // tells it what characters separate each word.  Control characters are
    // always placed in the table.  If no explicit characters are given in the
    // third argument then all space and punctuation characters are added by
    // default.  Otherwise just the given characters are added.
    //------------------------------------------------------------------------
    {
	int i;

	//-- Initialize the table with just control characters.
	for ( i = 0; i < 256; ++ i ) seps[ i ] = iscntrl( i );

	if ( argc > 1 && argv[ 1 ] ) {
	    char * p;

	    //-- Add the characters from the given arguments.
	    p = argv[ 1 ];
	    while ( * p ) seps[ 255 & * p ++ ] = 1;

	} else {

	    //-- Add space and punctuation characters.
	    for ( i = 0; i < 256; ++ i ) {
		if ( ispunct( i ) || isspace( i ) ) seps[ i ] = 1;
	    }
	}
    }

    //------------------------------
    // Initialize both binary trees.
    //------------------------------
    avl_init( & by_word, struct node, word, cmp_str );
    avl_init( & by_count, struct node, count, cmp_ul );

    //-- Initialize the struct used to find stuff in the tree.
    key_word.key = word;

    reads = 0;

    //-- Read each word from standard input.
    while ( 0 == readword( word, WORDSIZE, seps ) ) {

	++ reads;

	//-- Search for the word to see if it already exists.
	node_p = avl_find_exact( & by_word, & key_word );
	if ( node_p ) {

	    //-- Node found: increment count (which is also a key used later).
	    node_p->count.key ++;

	} else {

	    //-- Node not found: create and insert a new one
	    len = sizeof (struct node) - 16 + strlen( word ) + 1;
	    node_p = malloc( len );
	    if ( ! node_p ) {
		fprintf(stderr,"Out of memory (needed %lu bytes after %lu reads).\n",(unsigned long)len,reads);
		return 1;
	    }

	    //-- Initialize count (which is also a key used later).
	    node_p->count.key = 1;

	    //-- Copy word into the node (used now as a key).
	    strcpy( ( node_p->word.key = node_p->data ), word );

	    //-- Insert node into the tree that indexes by the word string.
	    avl_insert( & by_word, node_p );
	}
    }

    //-----------------------------------------------------------------------
    // Step through all the words in the "by word" tree and insert them into
    // the "by count" tree.  Duplicates have to be allowed because many words
    // may have the same count.
    //-----------------------------------------------------------------------
    words = 0;
    avl_loop_forward( & by_word, node_p ) {
	++ words;
	avl_insert_dup( & by_count, node_p );
    }

    //---------------------------------------------------------
    // Print all the words with count form the "by count" tree.
    //---------------------------------------------------------
    counts = 0;
    avl_loop_forward( & by_count, node_p ) {
	++ counts;
	if ( ! quiet ) printf( "%11lu %s\n", node_p->count.key, node_p->word.key );
    }

    printf( "%11lu total unique words\n",words);
    printf( "%11lu total counts\n",counts);

    return 0;
}
